/*
 * Jednoduche kresleni v Jave.
 * 
 * Jiri Skala, brezen 2008
 * 
 */
package simpleDrawing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import javax.swing.JOptionPane;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;

import javax.swing.event.MouseInputAdapter;

/**
 * Panel do ktereho budeme kreslit.
 * 
 * @author JSkala
 */
class PanelCanvas extends javax.swing.JPanel {
	
	/**
	 * Velikost krizku.
	 */
	private final int crossSize = 5;
	
	/**
	 * Tolerance pri chytani krizku mysi.
	 */
	private final int mouseTolerance = 7;
	
	
	
	
	
	/**
	 * Kliknutim se krizek smaze?
	 * Pokud ne, bude se posouvat.
	 */
	private boolean clickDeletes = false;
	
	/**
	 * Ma se kreslit tvar vyplneny obrazkem?
	 * Pokud ne, nakresli se obrazek cely.
	 */
	private boolean drawFilledShape = false;
	
	/**
	 * Seznam krizku.
	 */
	private ArrayList<Point> crosses;
	
	/**
	 * Index krizku, ktery tahneme.
	 * Hodnota -1 znamena, ze zadny netahneme.
	 */
	private int draggedCrossIndex = -1;
	
	/**
	 * Generator nahodnych cisel.
	 */
	private Random rnd;
	
	/**
	 * Rastrovy obrazek.
	 */
	private BufferedImage image = null;
	
	
	
	
	
	/**
	 * Konstruktor.
	 */
	public PanelCanvas() {
		
		// zavolame konstruktor panelu
		super();
		
		// zaregistrujeme obsluhu udalosti
		PanelMouseInputAdapter pmia = new PanelMouseInputAdapter();
		addMouseListener(pmia);
		addMouseMotionListener(pmia);
		
		// pripravime seznam
		crosses = new ArrayList<Point>();
		// pripravime generator nahodnych cisel
		rnd = new Random();
	}
	
	/**
	 * Vygeneruje spoustu krizku.
	 * Krizky pouze ulozi do seznamu, nic nekresli.
	 * 
	 * @param countX Pocet krizku v ose x.
	 * @param countY Pocet krizku v ose y.
	 */
	public void createCrosses(int countX, int countY) {
		
		// okraje
		final int margin = 10;
		
		// rozestupy mezi krizky
		float stepX = (this.getWidth() - 2*margin) / (float)(countX-1);
		float stepY = (this.getHeight() - 2*margin) / (float)(countY-1);
		
		// radky
		for (int row = 0; row < countY; row++) {
		
			int y = (int)(margin + stepY * row);
			
			// sloupce (jednotlive prvky radku)
			for (int column = 0; column < countX; column++) {
				
				int x = (int)(margin + stepX * column);
				
				crosses.add(new Point(x, y));
			}
		}
	}
	
	/**
	 * Nastavi zda se ma krizek pri kliknuti smazat.
	 * 
	 * @param del True pokud se ma krizek smazat, jinak false.
	 */
	public void setClickDeletes(boolean clickDeletes) {
		
		this.clickDeletes = clickDeletes;
	}
	
	/**
	 * Nastavi zda se ma kreslit tvar vyplneny obrazkem.
	 * 
	 * @param del True pokud se ma kreslit vyplneny tver, jinak false.
	 */
	public void setDrawFilledShape(boolean drawFilledShape) {
		this.drawFilledShape = drawFilledShape;
		repaint();
	}
	
	/**
	 * Metoda ktera se vyvola pri kresleni panelu.
	 * Nevolat primo, pouzivat repaint().
	 * 
	 * @param g Objekt pro kresleni.
	 */
	public void paint(Graphics g) {
	
		// pretypujeme Graphics na namaklejsi Graphics2D
		Graphics2D gr = (Graphics2D)g;
		
		// zjistime velikost panelu
		int w = this.getWidth();
		int h = this.getHeight();
		
		// nastavime barvu pozadi
		gr.setBackground(Color.white);
		
		// smazeme obsah panelu
		gr.clearRect(0, 0, w-1, h-1);
		
		if (image != null) {
			if ( ! drawFilledShape) {
				// nakreslime obrazek
				gr.drawImage(image, null, 0, 0);
			}
			else {
				// nastavime obrazek jako vypln
				Rectangle r = new Rectangle(0, 0, image.getWidth(), image.getHeight());
				gr.setPaint(new TexturePaint(image, r));
			
				// nakreslime neco
				gr.fillArc(10, 10, 300, 300, 45, 270);
			}
		}
		
		// nastavime barvu pro kresleni (obrys i vypln)
		gr.setColor(Color.black);
		// nastavime tlustejsi caru
		gr.setStroke(new BasicStroke(3));
		
		// nakreslime krizky
		for (int i = 0; i < crosses.size(); i++)
			drawCross(gr, crosses.get(i));
	}
	
	/**
	 * Posune krizek na novou pozici.
	 * 
	 * @param index	Index krizku, ktery se ma posunout.
	 * @param p		Souradnice kam se ma krizek posunout.
	 */
	private void moveCross(int index, Point p) {
		
		// vytvorime si graphics pro kresleni
		Graphics2D gr = (Graphics2D)getGraphics();
		
		// nastavime kresleni
		gr.setColor(Color.white);
		gr.setXORMode(Color.black);
		gr.setStroke(new BasicStroke(3));
		
		// smazeme puvodni krizek
		drawCross(gr, crosses.get(index));
		
		// nakreslime novy krizek
		drawCross(gr, p);
		
		// aktualizujeme souradnice krizku v seznamu
		crosses.set(index, p);
		
		// nezapomeneme graphics uvolnit!
		gr.dispose();
		
		// zadne repaint()!
	}
	
	/**
	 * Nakresli krizek.
	 * 
	 * @param gr	Graphics pro kresleni.
	 * @param p		Misto kam se ma krizek nakreslit.
	 */
	private void drawCross(Graphics2D gr, Point p) {
		
		int left = p.x - crossSize;
		int right = p.x + crossSize;
		int top = p.y - crossSize;
		int bottom = p.y + crossSize;
		
		gr.drawLine(left, top, right, bottom);
		gr.drawLine(left, bottom, right, top);
	}
	
	/**
	 * Zkusi najit krizek, ktery lezi na zadanem miste.
	 * 
	 * @param querry	Misto kde hledat.
	 * @return			Vraci index krizku, nebo -1 pokud na zadanem miste zadny krizek neni.
	 */
	private int findCross(Point querry) {
		
		// prohledame krizky
		for (int i = 0; i < crosses.size(); i++)
			// zjistime zda je nektery pobliz
			if (pointsClose(querry, crosses.get(i)))
				// nasli jsme!
				return i;
		
		// pokud jsme se dostali az sem, zadny krizek jsme nenasli
		return -1;
	}
	
	/**
	 * Zjisti zda jsou zadane dva body blizko u sebe.
	 * 
	 * @param pa	Prvni bod.
	 * @param pb	Druhy bod.
	 * @return		Vraci true pokud jsou zadane body blizko u sebe, jinak vraci false.
	 */
	private boolean pointsClose(Point pa, Point pb) {
		
		int dx = pa.x - pb.x;
		int dy = pa.y - pb.y;
		
		// nebudeme odmocnovat, zbytecne zdrzeni
		int dSqr = dx*dx + dy*dy;
		
		// misto odmocnovani porovname s druhou mocninou tolerance
		return dSqr <= mouseTolerance*mouseTolerance;
	}
	
	/**
	 * Nacte obrazek ze souboru.
	 * 
	 * @param inputFile Vstupni soubor.
	 */
	public void loadImage(File inputFile) {
		
		try {
			// nacteme obrazek
			image = ImageIO.read(inputFile);
			// prekreslime platno
			repaint();
		}
		catch (IOException exc) {
			// nacteni se nezdarilo
			image = null;
			JOptionPane.showMessageDialog(this, "Naten obrzku se nezdailo.",
					"Chyba", JOptionPane.ERROR_MESSAGE);
		}
	}
	
	/**
	 * Nakresli neco do obrazku.
	 */
	public void drawToImage() {
		
		if (image == null)
			// nemame zadny obrazek, nejaky si vytvorime
			image = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_3BYTE_BGR);
		
		// ziskame graphics pro kresleni
		Graphics2D gr = image.createGraphics();
		
		// zjistime velikost obrazku
		int w = image.getWidth();
		int h = image.getHeight();
		
		gr.setColor(Color.orange);
		gr.setStroke(new BasicStroke(5));
		gr.drawOval(0, 0, w-1, h-1);
		
		// nezapomeneme graphics uvolnit!
		gr.dispose();
		
		// zaplnime 5 % sumem
		for (int i = 0; i < w*h/20; i++) {
			
			// vygenerujeme nahodne souradnice
			int x = rnd.nextInt(w);
			int y = rnd.nextInt(h);
			
			// nastavime pixel
			image.setRGB(x, y, Color.magenta.getRGB());
		}
		
		// zobrazime vysledek
		repaint();
	}
	
	/**
	 * Ulozi obrazek jako PNG.
	 * 
	 * @param outputFile Vystupni soubor.
	 */
	public void saveImage(File outputFile) {

		try {
			// ulozime obrazek jako PNG
			ImageIO.write(image, "png", outputFile);
		}
		catch (IOException exc) {
			// ulozeni se nezdarilo
			JOptionPane.showMessageDialog(this, "Uloen obrzku se nezdailo.",
					"Chyba", JOptionPane.ERROR_MESSAGE);
		}
	}
	
	
	
	
	
	/**
	 * Trida pro obsluhu udalosti mysi na kreslicim platne.
	 * 
	 * @author Jiri Skala
	 */
	private class PanelMouseInputAdapter extends MouseInputAdapter {
		
		/**
		 * Reakce na kliknuti.
		 * 
		 * @param e Informace o udalosti.
		 */
		public void mouseClicked(MouseEvent e) {
			
			// kliknuto levym tlacitkem?
			if (e.getButton() == MouseEvent.BUTTON1) {
				// pridame krizek na soucasnou pozici mysi
				crosses.add(e.getPoint());
				// prekreslime
				repaint();
			}
			// kliknuto pravym tlacitkem?
			else if (e.getButton() == MouseEvent.BUTTON3) {
				int index;
				// mame smazat? je pobliz nejaky krizek?
				if (clickDeletes && (index = findCross(e.getPoint())) >= 0) {
					// smazeme krizek
					crosses.remove(index);
					// prekreslime
					repaint();
				}
			}
		}
		
		/**
		 * Reakce na stisknuti tlacitka mysi.
		 */
		public void mousePressed(MouseEvent e) {
			
			// chceme tahat?
			if (e.getButton() == MouseEvent.BUTTON3 && ! clickDeletes)
				// ano, zkusime najit krizek
				draggedCrossIndex = findCross(e.getPoint());
		}
		
		/**
		 * Reakce na uvolneni tlacitka mysi.
		 */
		public void mouseReleased(MouseEvent e) {
			
			// pokud se uvolnilo prave tlacitko
			if (e.getButton() == MouseEvent.BUTTON3) {
				// prestaneme tahat
				draggedCrossIndex = -1;
				repaint();
			}
		}
		
		/**
		 * Reakce na pohyb mysi SE STISKNUTYM TLACITKEM.
		 */
		public void mouseDragged(MouseEvent e) {
			
			// pokud tahneme nejaky krizek
			if (draggedCrossIndex >= 0) {
				// aktualizujeme jeho polohu
				moveCross(draggedCrossIndex, e.getPoint());
			}
		}
	}
}
